"""
utils test module
"""
import numpy as np

from nose import tools as nt

import tempfile, os #,shutil

from CFchecker import utils
from CFchecker.errors import CellMethodsParseError

class utils_test:
    #
    #   setup and teardown
    #
    #tst_base_dir = None
    #tst_subdir = 'test_dir1'
    #tst_base_files = ['foo', 'bar', 'baz']
    #tst_subdir_files = ['abc', 'def']
#
    #@classmethod
    #def setup_class(cls):
        #cls.tst_base_dir = tempfile.mkdtemp()
        #os.mkdir(os.path.join(cls.tst_base_dir, cls.tst_subdir))
        #def create_file(filename, cls=cls):
            #f = open(os.path.join(cls.tst_base_dir, filename), 'w')
            #f.write('foobarbaz\n')
            #f.close()
        #for filename in cls.tst_base_files:
            #create_file(filename)
        #for filename in cls.tst_subdir_files:
            #create_file(os.path.join(cls.tst_subdir, filename))
#
    #@classmethod
    #def teardown_class(cls):
        #shutil.rmtree(cls.tst_base_dir)

    def test_list_files(self):
        '''utils: list_files()'''
        def make_file(filename):
            f = open(filename, 'w')
            f.close()
        # create test directory structure
        tmpdir = tempfile.mkdtemp()
        level1_dir = os.path.join(tmpdir, 'level1_dir')
        level2_dir = os.path.join(tmpdir, 'level1_dir', 'level2_dir')
        os.mkdir(level1_dir)
        os.mkdir(level2_dir)
        make_file(os.path.join(tmpdir, 'level0_file1'))
        make_file(os.path.join(level1_dir, 'level1_file1'))
        make_file(os.path.join(level1_dir, 'level1_file2'))
        make_file(os.path.join(level2_dir, 'level2_file1'))

        # test with directory
        res = list(utils.list_files(tmpdir))
        expected = [ (tmpdir, set(['level1_dir']), set(['level0_file1'])), (level1_dir, set(['level2_dir']), set(['level1_file2', 'level1_file1'])),
                             (level2_dir, set([]), set(['level2_file1']))]
        nt.assert_equals(len(res), len(expected))
        for res_e, expected_e in zip(res, expected):
            nt.assert_equals( (res_e[0], set(res_e[1]), set(res_e[2])), expected_e)
        # test with file
        res = list(utils.list_files(os.path.join(tmpdir, 'level0_file1')))
        expected = [ (tmpdir, set([]), set(['level0_file1']))]
        nt.assert_equals(len(res), len(expected))
        for res_e, expected_e in zip(res, expected):
            nt.assert_equals( (res_e[0], set(res_e[1]), set(res_e[2])), expected_e)
        # test with nonexistent file/dir
        try:
            res = list(utils.list_files(os.path.join(tmpdir, 'level0_fileX')))
            raise AssertionError('OSError not raised')
        except OSError:
            pass

    def test_is_string_type(self):
        '''utils: is_string_type()'''
        string_types = ['foobar', '']
        for item in string_types:
            nt.assert_true(utils.is_string_type(item))

        non_strings = [5, 42.5, {}, [], ['string in list'], set()]
        for item in non_strings:
            nt.assert_false(utils.is_string_type(item))

    def test_is_num_type(self):
        '''utils: is_num_type()'''
        num_types = [42, 23.5, 1e-5, np.array(23), np.array([5])]
        for item in num_types:
            nt.assert_true(utils.is_num_type(item), msg=repr(item))

        non_nums = [{}, [], ['string in list'], 'string', set(), [3,4,6], np.array([23,5])]
        for item in non_nums:
            nt.assert_false(utils.is_num_type(item))

    def test_flatten(self):
        '''utils: flatten()'''
        a = []
        for i in range(100):
          a = [a, i]
        res = utils.flatten(a)
        expected = list(range(100))
        nt.assert_equals(res, expected)

    def test_find_duplicates(self):
        '''utils: find_duplicates()'''
        # no duplicates to be found
        items = ['lat', 'lon', 'time', 'lev']
        nt.assert_equals(utils.find_duplicates(items), [])
        # one duplicate to be found
        items = ['lat', 'lon', 'time', 'lon', 'lev']
        nt.assert_equals(set(utils.find_duplicates(items)), set(['lon']))
        # two duplicates to be found
        items = ['lat', 'lon', 'time', 'lon', 'lev', 'lat']
        nt.assert_equals(set(utils.find_duplicates(items)), set(['lon', 'lat']))

    def test_is_hashable(self):
        '''utils: is_hashable()'''
        hashable_items = ['abc', 23, 23.5, (1,2,3), frozenset([4,5,6])]
        for item in hashable_items:
            nt.assert_true(utils.is_hashable(item))
        unhashable_items = [[1,2,3], set([4,5,6]), {'foo': 3}, np.array([2,3,4]), np.array(5)]
        for item in unhashable_items:
            nt.assert_false(utils.is_hashable(item))

    def test_find_dim_vars(self):
        '''utils: find_dim_vars()'''
        class nCDFMock_Variable:
            dimensions = None
            def __init__(self,dimensions, axis=None):
                self.dimensions = dimensions
                if axis: self.axis = axis

        class nCDFMock_File:
            dimensions = {'lat':5, 'lon': 3, 'time1': 10, 'time2': 12}
            variables = {'lat': nCDFMock_Variable(('lat',), axis='Y'), 'lon': nCDFMock_Variable(('lon',), axis='X'),
                                'time1': nCDFMock_Variable(('time1',), axis='T'), 'time2': nCDFMock_Variable(('time2',), axis='T'),
                                'temp': nCDFMock_Variable(('lat','lon'))}

        nt.assert_equals(set(utils.find_dim_vars(nCDFMock_File())), set(['lat','lon', 'time1', 'time2']))

    def test_map_axes(self):
        '''utils: map_axes()'''
        class nCDFMock_Variable:
            def __init__(self,dimensions, axis=None):
                self.dimensions = dimensions
                if axis: self.axis = axis

        class nCDFMock_File:
            dimensions = {'lat':5, 'lon': 3, 'time1': 10, 'time2': 12}
            variables = {'lat': nCDFMock_Variable(('lat',), axis='Y'), 'lon': nCDFMock_Variable(('lon',), axis='X'),
                                'time1': nCDFMock_Variable(('time1',), axis='T'), 'time2': nCDFMock_Variable(('time2',), axis='T'),
                                'temp': nCDFMock_Variable(('lat','lon'))}

        # reverse_map = False
        # as list order is not predictable for reverse_map=False, a by-element approach is needed
        expected = {'X': ['lon'], 'Y': ['lat'], 'T': ['time1', 'time2']}
        res = utils.map_axes(nCDFMock_File(), reverse_map=False)
        nt.assert_equals(set(res.keys()), set(expected.keys()))
        for k in expected.keys():
            nt.assert_equals(set(res[k]), set(expected[k]))

        # reverse_map = True
        expected = {'lon': 'X', 'lat': 'Y', 'time1': 'T', 'time2': 'T'}
        res = utils.map_axes(nCDFMock_File(), reverse_map=True)
        nt.assert_equals(res, expected)

    def test_map_dim_vars_to_types(self):
        '''utils: map_dim_vars_to_types()'''
        class nCDFMock_Variable:
            def __init__(self, units=None,dimensions=None,standard_name=None, long_name=None,axis=None,
                                    positive=None):
                if dimensions: self.dimensions = dimensions
                if units: self.units = units
                if standard_name: self.standard_name = standard_name
                if long_name: self.long_name = long_name
                if axis: self.axis = axis
                if positive: self.positive = positive

        class nCDFMock_File:
            dimensions = {'lat': 5, 'lat2': 3, 'lon': 4, 'time1': 2, 'lev': 3, 'time2': 4}
            variables = {'lat': nCDFMock_Variable(axis='Y',dimensions=('lat',)), # identify by axis
                                'lat2': nCDFMock_Variable(dimensions=('lat2',), units='degrees_north'), # identify by units
                                'lon': nCDFMock_Variable(dimensions=('lon',), standard_name='longitude'), # itdntifiy by standard_name
                                'time1': nCDFMock_Variable( units='days since 2011-01-01', dimensions=('time1',)), # identify by units
                                'lev': nCDFMock_Variable(dimensions=('lev',), positive='up'), # identify by positive
                                'time2': nCDFMock_Variable(dimensions=('time2',)), # not identified
                                }

        var_map = utils.map_dim_vars_to_types(nCDFMock_File())
        expected = {'lat': 'Y', 'lat2': 'Y', 'lon': 'X', 'lev': 'Z', 'time1': 'T', 'time2': None}
        nt.assert_equals(var_map, expected)
        var_map = utils.map_dim_vars_to_types(nCDFMock_File(), reverse_map=True)
        expected = {'Y': ['lat', 'lat2'], 'X': ['lon'], 'Z': ['lev'],'T': ['time1'], None: ['time2']}
        # as list order is not predictable, a by-element approach is needed
        nt.assert_equals(set(var_map.keys()), set(expected.keys()))
        for k in expected.keys():
            nt.assert_equals(set(var_map[k]), set(expected[k]))

    def test_find_bounds(self):
        '''utils: find_bounds()'''
        class nCDFMock_Variable:
            def __init__(self,bounds=None):
                if bounds: self.bounds = bounds

        class nCDFMock_File:
            variables = {'lat': nCDFMock_Variable(bounds='lat_bnds'), 'lon': nCDFMock_Variable(bounds='lon_bnds'),
                                'time1': nCDFMock_Variable(bounds='time1_bnds'), 'time2': nCDFMock_Variable(),
                                'lat_bnds': nCDFMock_Variable(), 'lon_bounds': nCDFMock_Variable(), # wrong spelling for lon_bounds
                                'time1_bnds': nCDFMock_Variable(), 'time2_bnds': nCDFMock_Variable(),
                                'temp': nCDFMock_Variable()}

        expected = {'lat': 'lat_bnds', 'time1': 'time1_bnds'}
        res = utils.find_bounds(nCDFMock_File())
        nt.assert_equals(res, expected)

    def test_is_valid_cf_name(self):
        '''utils: is_valid_cf_name()'''
        names_ok = ['vmr_o3', 'time', 'lat', 'lon', 'title']
        for name in names_ok:
            nt.assert_true(utils.is_valid_cf_name(name))
        names_bad = ['12345', '1time', '_FillValue', 'lat-lon', '#comment', u'f\xc3\xbcbar']
        for name in names_bad:
            nt.assert_false(utils.is_valid_cf_name(name))

    def test_parse_cell_methods(self):
        '''utils: parse_cell_methods()'''
        cell_method_strings_ok = ['',
                                  # the following strings are the examples taken from the CF-1.0 conventions document
                                  'time: maximum', 'lon: maximum time: mean', 'time: mean lon: maximum',
                                  'lat: lon: standard_deviation', 'lon: standard_deviation lat: standard_deviation',
                                  'time: standard_deviation (interval: 1 day)', 'time: standard_deviation (interval: 1 year)',
                                  'lat: lon: standard_deviation (interval: 10 km)',
                                  'lat: lon: standard_deviation (interval: 0.1 degree_N interval: 0.2 degree_E)',
                                  'lat: mean (area-weighted)', 'lat: mean (interval: 1 degree_north comment: area-weighted)',
                                  'time: minimum within years time: mean over years',
                                  'time: MaxiMum', # case sensitive test for methods
                                  'area: mean where land_sea',   # where example
                                  'area: mean where land_sea over sea', # where ... over
                                  'area: mean where land_sea over sea over days', # where ... over ... over
                                  'area: mean where land_sea over sea within years', # where ... over ... within
                                  'area: mean where land_sea within days', # where ... within
                                  ]
        cell_method_parse_results_ok = [ [],
                                         [(['time'], 'maximum', '', '')], [(['lon'], 'maximum', '', ''), (['time'], 'mean', '', '')],
                                         [(['time'], 'mean', '', ''), (['lon'], 'maximum', '', '')], [(['lat', 'lon'],
                                                                                               'standard_deviation', '', '')],
                                         [(['lon'], 'standard_deviation', '', ''), (['lat'], 'standard_deviation', '', '')],
                                         [(['time'], 'standard_deviation', '', 'interval: 1 day')], [(['time'], 'standard_deviation',
                                                                                                  '', 'interval: 1 year')],
                                         [(['lat', 'lon'], 'standard_deviation', '', 'interval: 10 km')],
                                         [(['lat', 'lon'], 'standard_deviation', '', 'interval: 0.1 degree_N interval: 0.2 degree_E')],
                                         [(['lat'], 'mean', '', 'area-weighted')], [(['lat'], 'mean', '',
                                                                                 'interval: 1 degree_north comment: area-weighted')],
                                         [(['time'], 'minimum',  'within years', ''), (['time'], 'mean', 'over years', '')],
                                         [(['time'], 'maximum', '', '')],
                                         [(['area'], 'mean', 'where land_sea', '')],
                                         [(['area'], 'mean', 'where land_sea over sea', '')],
                                         [(['area'], 'mean', 'where land_sea over sea over days', '')],
                                         [(['area'], 'mean', 'where land_sea over sea within years', '')],
                                         [(['area'], 'mean', 'where land_sea within days', '')],
                                       ]
        for elem, res in zip(cell_method_strings_ok, cell_method_parse_results_ok):
            nt.assert_equal(utils.parse_cell_methods(elem), res)
        cell_method_strings_bad = [':', # no name given
                                   'time maximum', # separator after name missing
                                   'time:', # method missing
                                   'lat: ', # method missing with space after ":"
                                   'lon: 123foo: point', # second name invalid
                                   'lon: lat:', # method missing after 2 names
                                   'time: spam', # invalid method
                                   'time: sum (area-weighted', # closing parenthesis missing with only one extra_info chunk
                                   'time: sum (area-weighted interval: 1 degree_north', # closing parenthesis missing with more than one chunk
                                   'time: sum (area-weighted (comment: interval is 1 degree_north)', # opening parenthesis within extra_info
                                   'time: minimum within blubs time: mean over years',  # invalid climatological units
                                   'time: minimum within years time: mean withover years',   # expected name or climatological specifier after "mean"
                                   'time: 123invalid: point',    # 2nd name invalid
                                   '123time: time: point',    # 1st name invalid
                                   'time: sum 123invalid: point',    # name for 2nd method invalid
                                   ]
        for elem in cell_method_strings_bad:
            nt.assert_raises(CellMethodsParseError, utils.parse_cell_methods, elem)

    def test_parse_cell_methods_where_statement(self):
        '''utils: parse_cell_methods_where_statement()'''
        statement_strings = [ '', 'within years', 'over years', 'where land_sea', 'where land_sea over sea',
                              'where land_sea over sea over days', 'where land_sea over sea within years',
                              'where land_sea within days']
        parse_results = [ (None, None), (None, None), (None, None), ('land_sea', None), ('land_sea', 'sea'),
                          ('land_sea', 'sea'), ('land_sea', 'sea'), ('land_sea', None) ]
        for idx, (elem, exp) in enumerate(zip(statement_strings, parse_results)):
            res = utils.parse_cell_methods_where_statement(elem)
            nt.assert_equal(res, exp, msg='%d: %s != %s' % (idx, res, exp))

    def test_parse_cell_methods_climatology_statement(self):
        '''utils: parse_cell_methods_climatology_statement()'''
        statement_strings = [ '', 'within years', 'over years', 'where land_sea', 'where land_sea over sea',
                              'where land_sea over sea over days', 'where land_sea over sea within years',
                              'where land_sea within days']
        parse_results = [ (None, None), ('within', 'years'), ('over', 'years'), (None, None), (None, None),
                          ('over', 'days'), ('within', 'years'), ('within', 'days') ]
        for idx, (elem, exp) in enumerate(zip(statement_strings, parse_results)):
            res = utils.parse_cell_methods_climatology_statement(elem)
            nt.assert_equal(res, exp, msg='%d: %s != %s' % (idx, res, exp))

    def test_find_climatology_vars(self):
        '''utils: find_climatology_vars()'''
        class nCDFMock_Variable:
            def __init__(self,bounds=None,climatology=None, units=None,dimensions=None,cell_methods=None):
                if dimensions: self.dimensions = dimensions
                if bounds: self.bounds = bounds
                if climatology: self.climatology = climatology
                if units: self.units = units
                if cell_methods: self.cell_methods = cell_methods

        class nCDFMock_File:
            dimensions = {'lat': 5, 'lon': 4, 'time1': 2, 'time2': 3, 'time3': 6, 'time4': 7, 'bnds': 2}
            variables = {'lat': nCDFMock_Variable(bounds='lat_bnds',units='degrees_north',dimensions=('lat',)),
                                'lon': nCDFMock_Variable(bounds='lon_bnds',units='degrees_east', dimensions=('lon',)),
                                'time1': nCDFMock_Variable(bounds='time1_bnds', units='days since 2011-01-01', dimensions=('time1',)),
                                'time2': nCDFMock_Variable(climatology='climatology_bnds', dimensions=('time2',)),
                                'time3': nCDFMock_Variable(units='days since 0-1-1', dimensions=('time3',)),
                                'time4': nCDFMock_Variable(units='days since 2011-01-01', dimensions=('time4',)),
                                'lat_bnds': nCDFMock_Variable(dimensions=('lat', 'bnds')), 'lon_bnds': nCDFMock_Variable(dimensions=('lon', 'bnds')),
                                'time1_bnds': nCDFMock_Variable(dimensions=('time1', 'bnds')),
                                'climatology_bnds': nCDFMock_Variable(dimensions=('time2', 'bnds')),
                                'o3': nCDFMock_Variable(dimensions=('lat', 'lon','time1')),
                                'temp': nCDFMock_Variable(dimensions=('lat', 'lon','time2')),
                                'o2': nCDFMock_Variable(dimensions=('lat', 'lon','time3')),
                                'o5': nCDFMock_Variable(dimensions=('lat', 'lon','time4'), cell_methods='time4: sum within years time4: mean over years'),
                                }

        expected = set(['climatology_bnds', 'temp', 'o2', 'o5'])
        res = set(utils.find_climatology_vars(nCDFMock_File()))
        nt.assert_equals(res, expected)

    def test_find_compress_vars(self):
        '''utils: find_compress_vars()'''
        class nCDFMock_Variable:
            def __init__(self,dimensions=None,compress=None,dtype=None):
                if dimensions: self.dimensions = dimensions
                if compress: self.compress = compress
                if dtype is not None: self.dtype = dtype    # bool(dtype) is False

        class nCDFMock_File:
            dimensions = {'lat': 5, 'lon': 4, 'time': 2, 'compressed_lat_lon': 3}
            variables = {'lat': nCDFMock_Variable(dimensions=('lat',), dtype=np.dtype(np.float)),
                                'lon': nCDFMock_Variable(dimensions=('lon',), dtype=np.dtype(np.float)),
                                'compressed_lat_lon': nCDFMock_Variable(dimensions=('compressed_lat_lon',), compress='lat not_there lon', dtype=np.dtype(np.int32)),
                                'time': nCDFMock_Variable(dimensions=('time',), dtype=np.dtype(np.float)),
                                'o3': nCDFMock_Variable(dimensions=('lat', 'lon','time'),dtype=np.dtype(np.float)),
                                }

        expected = {'compressed_lat_lon': ['lat', 'lon']}
        res = utils.find_compress_vars(nCDFMock_File())
        nt.assert_equals(res, expected)

    def test_find_auxiliary_coord_vars(self):
        '''utils: find_auxiliary_coord_vars()'''
        class nCDFMock_Variable:
            def __init__(self,dimensions=None,coordinates=None):
                if dimensions is not None: self.dimensions = dimensions
                if coordinates is not None: self.coordinates = coordinates

        class nCDFMock_File:
            dimensions = {'lat': 5, 'lon': 4, 'time': 2, 'xc': 3, 'yc': 3}
            variables = {'lat': nCDFMock_Variable(dimensions=('lat',),),
                                'lon': nCDFMock_Variable(dimensions=('lon',)),
                                'xc': nCDFMock_Variable(dimensions=('xc',)),
                                'yc': nCDFMock_Variable(dimensions=('yc',)),
                                'lon2d': nCDFMock_Variable(dimensions=('xc','yc')),
                                'lat2d': nCDFMock_Variable(dimensions=('xc','yc')),
                                'time': nCDFMock_Variable(dimensions=('time','xc','yc'), coordinates='lon2d lat2d does_not_exist')
                                }

        expected = {'time': ['lon2d', 'lat2d']}
        res = utils.find_auxiliary_coord_vars(nCDFMock_File())
        nt.assert_equals(res, expected)

    def test_find_scalar_coord_vars(self):
        '''utils: find_scalar_coord_vars()'''
        class nCDFMock_Variable:
            def __init__(self,dimensions=None,coordinates=None):
                if dimensions is not None: self.dimensions = dimensions
                if coordinates is not None: self.coordinates = coordinates

        class nCDFMock_File:
            dimensions = {'lat': 5, 'lon': 4, 'time': 2, 'xc': 3, 'yc': 3}
            variables = {'lat': nCDFMock_Variable(dimensions=('lat',),),
                                'lon': nCDFMock_Variable(dimensions=('lon',)),
                                'xc': nCDFMock_Variable(dimensions=('xc',)),
                                'yc': nCDFMock_Variable(dimensions=('yc',)),
                                'lon2d': nCDFMock_Variable(dimensions=('xc','yc')),
                                'lat2d': nCDFMock_Variable(dimensions=('xc','yc')),
                                'scalar_var': nCDFMock_Variable(dimensions=()),
                                'time': nCDFMock_Variable(dimensions=('time','xc','yc'), coordinates='lon2d lat2d scalar_var')
                                }

        expected = {'time': ['scalar_var']}
        res = utils.find_scalar_coord_vars(nCDFMock_File())
        nt.assert_equals(res, expected)

    def test_find_label_vars(self):
        '''utils: find_label_vars()'''
        class dtypeMock:
            def __init__(self,dtypechar):
                self.char = dtypechar

        class nCDFMock_Variable:
            def __init__(self,dimensions=None,dtypechar=None, standard_name=None, coordinates=None):
                if dimensions is not None: self.dimensions = dimensions; self.ndim = len(dimensions)
                if dtypechar is not None: self.dtype = dtypeMock(dtypechar)
                if standard_name is not None: self.standard_name = standard_name
                if coordinates is not None: self.coordinates = coordinates

        class nCDFMock_File:
            dimensions = {'lat': 5, 'lon': 4, 'time': 2, 'station': 2, 'str_len': 5, 'areas': 4}
            variables = {'lat': nCDFMock_Variable(dimensions=('lat',),dtypechar='f',standard_name='latitude'),
                                'lon': nCDFMock_Variable(dimensions=('lon',),dtypechar='f', standard_name='llongitude'),
                                'time': nCDFMock_Variable(dimensions=('time',),dtypechar='f', standard_name='time'),
                                'station': nCDFMock_Variable(dimensions=('station', 'str_len'), dtypechar='S', standard_name='station_names'),
                                'areas': nCDFMock_Variable(dimensions=('areas', 'str_len'), dtypechar='S', standard_name='area_type'),
                                'temp': nCDFMock_Variable(dimensions=('time', 'lat', 'lon', 'station'),dtypechar='f', standard_name='temperature',
                                                                          coordinates='station areas'),
                                }
        expected = ['station']
        res = utils.find_label_vars(nCDFMock_File())
        nt.assert_equals(res, expected)

    def test_find_grid_mapping_vars(self):
        '''utils: find_grid_mapping_vars()'''
        class nCDFMock_Variable:
            def __init__(self,dimensions=None,grid_mapping=None):
                if dimensions is not None: self.dimensions = dimensions; self.ndim = len(dimensions)
                if grid_mapping is not None: self.grid_mapping = grid_mapping

        class nCDFMock_File:
            dimensions = {'lat': 5, 'lon': 4, 'time': 2}
            variables = {'lat': nCDFMock_Variable(dimensions=('lat',)),
                                'lon': nCDFMock_Variable(dimensions=('lon',)),
                                'time': nCDFMock_Variable(dimensions=('time',)),
                                'grid_mapping_var': nCDFMock_Variable(dimensions=(),),
                                'temp': nCDFMock_Variable(dimensions=('time', 'lat', 'lon'), grid_mapping='grid_mapping_var'),
                                'temp2': nCDFMock_Variable(dimensions=('time', 'lat', 'lon'), grid_mapping='var_that_does_not_exist'),
                                }
        expected = {'temp': 'grid_mapping_var'}
        res = utils.find_grid_mapping_vars(nCDFMock_File())
        nt.assert_equals(res, expected)

    def test_find_area_type_vars(self):
        '''utils: find_area_type_vars()'''
        class dtypeMock:
            def __init__(self,dtypechar):
                self.char = dtypechar

        class nCDFMock_Variable:
            def __init__(self,dimensions=None, standard_name=None, dtypechar=None):
                if dimensions is not None: self.dimensions = dimensions
                if standard_name is not None: self.standard_name = standard_name
                if dtypechar is not None: self.dtype = dtypeMock(dtypechar)

        class nCDFMock_Variable:
            def __init__(self,dimensions=None,dtypechar=None, standard_name=None, coordinates=None):
                if dimensions is not None: self.dimensions = dimensions; self.ndim = len(dimensions)
                if dtypechar is not None: self.dtype = dtypeMock(dtypechar)
                if standard_name is not None: self.standard_name = standard_name
                if coordinates is not None: self.coordinates = coordinates

        class nCDFMock_File:
            dimensions = {'lat': 5, 'lon': 4, 'time': 2, 'station': 2, 'str_len': 5, 'areas': 4}
            variables = {'lat': nCDFMock_Variable(dimensions=('lat',),dtypechar='f',standard_name='latitude'),
                                'lon': nCDFMock_Variable(dimensions=('lon',),dtypechar='f', standard_name='llongitude'),
                                'time': nCDFMock_Variable(dimensions=('time',),dtypechar='f', standard_name='time'),
                                'station': nCDFMock_Variable(dimensions=('station', 'str_len'), dtypechar='S', standard_name='station_names'),
                                'areas': nCDFMock_Variable(dimensions=('areas', 'str_len'), dtypechar='S', standard_name='area_type'),
                                'temp': nCDFMock_Variable(dimensions=('time', 'lat', 'lon', 'station'),dtypechar='f', standard_name='temperature',
                                                                          coordinates='station areas'),
                                }

        expected = ['areas']
        res = utils.find_area_type_vars(nCDFMock_File())
        nt.assert_equals(res, expected)

    def test_get_cf_version(self):
        '''utils: get_cf_version()'''
        class nCDFFileMock_with_Att:
            Conventions = "CF-1.x"
        class nCDFFileMock_no_Att:
            other_attribute = "spam"
        nt.assert_equals(utils.get_cf_version(nCDFFileMock_with_Att()), "1.x")
        nt.assert_equals(utils.get_cf_version(nCDFFileMock_no_Att()), None)


    def test_units_known(self):
        '''utils: units_known()'''
        nt.assert_false(utils.units_known(23.5))

        units = ['K', 'm', 'second', 'm2', 's-3', 'days since 2000-01-01']
        for unit in units:
            nt.assert_true(utils.units_known(unit))
        units =  ['foo', 'barr', 'baz', 23, None]
        for unit in units:
            nt.assert_false(utils.units_known(unit))

    def test_units_convertible(self):
        '''utils: units_convertible()'''
        units = [('K', 'Celsius'), ('m', 'in'), ('liter', 'gallon'), ('litre', 'liter'),
                 ('m2', 'cm2'), ('days since 2010-01-01', 'seconds since 1970-01-01'),
                 ('hours', 'days'), ('days since 1970-01-01', 's'), ('degrees_north', '1'),
                 ('degrees_east', 'degrees_north'), ('degreesN', 'degrees')]
        for unit1, unit2 in units:
            nt.assert_true(utils.units_convertible(unit1, unit2, reftimeistime=True))
        # contains invalid units
        inv_units = [('K', 'Spam'), ('cat', 'dog')]
        # valid units but not convertible
        unconv_units = [('K', 'liter'), ('m', 'second'), ('liter', 'light_year'),
                        ('faraday', 'liter'), ('m3', 's3')]
        for unit1, unit2 in inv_units:
            nt.assert_raises(utils.UnknownUnitsError, utils.units_convertible, unit1, unit2, True)
        for unit1, unit2 in unconv_units:
            nt.assert_false(utils.units_convertible(unit1, unit2, reftimeistime=True))
        # reftimeistime = False
        units = [('K', 'Celsius'), ('m', 'in'), ('liter', 'gallon'), ('litre', 'liter'),
                 ('m2', 'cm2'), ('days since 2010-01-01', 'seconds since 1970-01-01'),
                 ('hours', 'days'), ('degrees_north', '1'),
                 ('degrees_east', 'degrees_north'), ('degreesN', 'degrees')]
        for unit1, unit2 in units:
            nt.assert_true(utils.units_convertible(unit1, unit2, reftimeistime=False))

        units = [('days since 1970-01-01', 's'),]
        for unit1, unit2 in units:
            nt.assert_false(utils.units_convertible(unit1, unit2, reftimeistime=False))

    def test_comp_asc(self):
        '''utils: comp_asc()'''
        vals = [(1,2), (2,1), ('a', 'b'), (23, 23), ('abc', 'abc'), ('1.0.2', '1.0.3'), ('1.2.1', '1.2')]
        expected = [-1, 1, -1, 0, 0, -1, 1]
        for (x,y), exp in zip(vals, expected):
            res = utils.comp_asc(x,y)
            nt.assert_equals(res, exp)

    def test_comp_desc(self):
        '''utils: comp_desc()'''
        vals = [(1,2), (2,1), ('a', 'b'), (23, 23), ('abc', 'abc'), ('1.0.2', '1.0.3'), ('1.2.1', '1.2')]
        expected = [1, -1, 1, 0, 0, 1, -1]
        for (x,y), exp in zip(vals, expected):
            res = utils.comp_desc(x,y)
            nt.assert_equals(res, exp)

    def test_is_ascending(self):
        '''utils: is_ascending()'''
        inputs = [[1,2,3,4,5,6,7,8,9], [1,2,4,5,5,6], [1,5,8,12,34,11],
                  np.array([34,56,67,867]), np.array([34,56,98,123,453,453,600]),
                  np.array([34,56,23,75,23])]
        expected = [True, False, False, True, False, False]
        for inp, exp in zip(inputs, expected):
            res = utils.is_ascending(inp)
            nt.assert_equals(res, exp)

    def test_is_monotonic(self):
        '''utils: is_monotonic()'''
        inputs = [[1,2,3,4,5,6,7,8,9], [1,2,4,5,5,6], [1,5,8,12,34,11],
                  [6,4,3,2,1], [9,8,7,6,5,5,3],
                  np.array([34,56,67,867]), np.array([34,56,98,123,453,453,600]),
                  np.array([34,56,23,75,23]), np.array([5,4,3,1]), np.array([99,88,77,66,66,3])]
        expected = [True, False, False, True, False, True, False, False, True, False]
        for inp, exp in zip(inputs, expected):
            res = utils.is_monotonic(inp, circular=False)
            nt.assert_equals(res, exp)
        inputs = [[1,2,3,4,5,6,7,8,9], [4,5,6,7,7,1,2], [4,5,6,7,1,1,2], [5,5,6,1,2,4],
                  [6,4,3,2,1], [6,5,3,9,8,7], np.array([123,453,600,55,34,56,98,])]
        expected = [True, False, False, False, True, True, False]
        for inp, exp in zip(inputs, expected):
            res = utils.is_monotonic(inp, circular=True)
            nt.assert_equals(res, exp)
